# Copyright 2021 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# pylint: disable=g-bad-import-order,redefined-builtin
"""APIs to create the searcher model.

Task guide:
https://www.tensorflow.org/lite/tutorials/model_maker_text_searcher.
"""

from tensorflow_examples.lite.model_maker.core.data_util.image_searcher_dataloader import DataLoader as ImageDataLoader
from tensorflow_examples.lite.model_maker.core.data_util.metadata_loader import MetadataType
from tensorflow_examples.lite.model_maker.core.data_util.searcher_dataloader import DataLoader
from tensorflow_examples.lite.model_maker.core.data_util.text_searcher_dataloader import DataLoader as TextDataLoader
from tensorflow_examples.lite.model_maker.core.task.searcher import ExportFormat
from tensorflow_examples.lite.model_maker.core.task.searcher import ScaNNOptions
from tensorflow_examples.lite.model_maker.core.task.searcher import ScoreAH
from tensorflow_examples.lite.model_maker.core.task.searcher import ScoreBruteForce
from tensorflow_examples.lite.model_maker.core.task.searcher import Searcher
from tensorflow_examples.lite.model_maker.core.task.searcher import Tree



# THIS IS A GENERATED FILE. DO NOT EDIT!
# pylint: skip-file
# pyformat: disable
# Copyright 2021 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: input-config.proto

from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='input-config.proto',
  package='tensorflow_examples.lite.examples.recommendation.ml_v2.configs',
  syntax='proto2',
  serialized_options=None,
  serialized_pb=b'\n\x12input-config.proto\x12>tensorflow_examples.lite.examples.recommendation.ml_v2.configs\"\xd9\x01\n\x07\x46\x65\x61ture\x12\x14\n\x0c\x66\x65\x61ture_name\x18\x01 \x01(\t\x12\x61\n\x0c\x66\x65\x61ture_type\x18\x02 \x01(\x0e\x32K.tensorflow_examples.lite.examples.recommendation.ml_v2.configs.FeatureType\x12\x12\n\nvocab_name\x18\x03 \x01(\t\x12\x12\n\nvocab_size\x18\x04 \x01(\x03\x12\x15\n\rembedding_dim\x18\x05 \x01(\x03\x12\x16\n\x0e\x66\x65\x61ture_length\x18\x06 \x01(\x03\"\xcc\x01\n\x0c\x46\x65\x61tureGroup\x12Y\n\x08\x66\x65\x61tures\x18\x01 \x03(\x0b\x32G.tensorflow_examples.lite.examples.recommendation.ml_v2.configs.Feature\x12\x61\n\x0c\x65ncoder_type\x18\x02 \x01(\x0e\x32K.tensorflow_examples.lite.examples.recommendation.ml_v2.configs.EncoderType\"\xc9\x02\n\x0bInputConfig\x12k\n\x15global_feature_groups\x18\x01 \x03(\x0b\x32L.tensorflow_examples.lite.examples.recommendation.ml_v2.configs.FeatureGroup\x12m\n\x17\x61\x63tivity_feature_groups\x18\x02 \x03(\x0b\x32L.tensorflow_examples.lite.examples.recommendation.ml_v2.configs.FeatureGroup\x12^\n\rlabel_feature\x18\x03 \x01(\x0b\x32G.tensorflow_examples.lite.examples.recommendation.ml_v2.configs.Feature*-\n\x0b\x46\x65\x61tureType\x12\n\n\x06STRING\x10\x00\x12\x07\n\x03INT\x10\x01\x12\t\n\x05\x46LOAT\x10\x02*)\n\x0b\x45ncoderType\x12\x07\n\x03\x42OW\x10\x00\x12\x07\n\x03\x43NN\x10\x01\x12\x08\n\x04LSTM\x10\x02'
)

_FEATURETYPE = _descriptor.EnumDescriptor(
  name='FeatureType',
  full_name='tensorflow_examples.lite.examples.recommendation.ml_v2.configs.FeatureType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='STRING', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='INT', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FLOAT', index=2, number=2,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=845,
  serialized_end=890,
)
_sym_db.RegisterEnumDescriptor(_FEATURETYPE)

FeatureType = enum_type_wrapper.EnumTypeWrapper(_FEATURETYPE)
_ENCODERTYPE = _descriptor.EnumDescriptor(
  name='EncoderType',
  full_name='tensorflow_examples.lite.examples.recommendation.ml_v2.configs.EncoderType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='BOW', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CNN', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LSTM', index=2, number=2,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=892,
  serialized_end=933,
)
_sym_db.RegisterEnumDescriptor(_ENCODERTYPE)

EncoderType = enum_type_wrapper.EnumTypeWrapper(_ENCODERTYPE)
STRING = 0
INT = 1
FLOAT = 2
BOW = 0
CNN = 1
LSTM = 2



_FEATURE = _descriptor.Descriptor(
  name='Feature',
  full_name='tensorflow_examples.lite.examples.recommendation.ml_v2.configs.Feature',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='feature_name', full_name='tensorflow_examples.lite.examples.recommendation.ml_v2.configs.Feature.feature_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='feature_type', full_name='tensorflow_examples.lite.examples.recommendation.ml_v2.configs.Feature.feature_type', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='vocab_name', full_name='tensorflow_examples.lite.examples.recommendation.ml_v2.configs.Feature.vocab_name', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='vocab_size', full_name='tensorflow_examples.lite.examples.recommendation.ml_v2.configs.Feature.vocab_size', index=3,
      number=4, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='embedding_dim', full_name='tensorflow_examples.lite.examples.recommendation.ml_v2.configs.Feature.embedding_dim', index=4,
      number=5, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='feature_length', full_name='tensorflow_examples.lite.examples.recommendation.ml_v2.configs.Feature.feature_length', index=5,
      number=6, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=87,
  serialized_end=304,
)


_FEATUREGROUP = _descriptor.Descriptor(
  name='FeatureGroup',
  full_name='tensorflow_examples.lite.examples.recommendation.ml_v2.configs.FeatureGroup',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='features', full_name='tensorflow_examples.lite.examples.recommendation.ml_v2.configs.FeatureGroup.features', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='encoder_type', full_name='tensorflow_examples.lite.examples.recommendation.ml_v2.configs.FeatureGroup.encoder_type', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=307,
  serialized_end=511,
)


_INPUTCONFIG = _descriptor.Descriptor(
  name='InputConfig',
  full_name='tensorflow_examples.lite.examples.recommendation.ml_v2.configs.InputConfig',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='global_feature_groups', full_name='tensorflow_examples.lite.examples.recommendation.ml_v2.configs.InputConfig.global_feature_groups', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='activity_feature_groups', full_name='tensorflow_examples.lite.examples.recommendation.ml_v2.configs.InputConfig.activity_feature_groups', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='label_feature', full_name='tensorflow_examples.lite.examples.recommendation.ml_v2.configs.InputConfig.label_feature', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=514,
  serialized_end=843,
)

_FEATURE.fields_by_name['feature_type'].enum_type = _FEATURETYPE
_FEATUREGROUP.fields_by_name['features'].message_type = _FEATURE
_FEATUREGROUP.fields_by_name['encoder_type'].enum_type = _ENCODERTYPE
_INPUTCONFIG.fields_by_name['global_feature_groups'].message_type = _FEATUREGROUP
_INPUTCONFIG.fields_by_name['activity_feature_groups'].message_type = _FEATUREGROUP
_INPUTCONFIG.fields_by_name['label_feature'].message_type = _FEATURE
DESCRIPTOR.message_types_by_name['Feature'] = _FEATURE
DESCRIPTOR.message_types_by_name['FeatureGroup'] = _FEATUREGROUP
DESCRIPTOR.message_types_by_name['InputConfig'] = _INPUTCONFIG
DESCRIPTOR.enum_types_by_name['FeatureType'] = _FEATURETYPE
DESCRIPTOR.enum_types_by_name['EncoderType'] = _ENCODERTYPE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Feature = _reflection.GeneratedProtocolMessageType('Feature', (_message.Message,), {
  'DESCRIPTOR' : _FEATURE,
  '__module__' : 'input_config_pb2'
  # @@protoc_insertion_point(class_scope:tensorflow_examples.lite.examples.recommendation.ml_v2.configs.Feature)
  })
_sym_db.RegisterMessage(Feature)

FeatureGroup = _reflection.GeneratedProtocolMessageType('FeatureGroup', (_message.Message,), {
  'DESCRIPTOR' : _FEATUREGROUP,
  '__module__' : 'input_config_pb2'
  # @@protoc_insertion_point(class_scope:tensorflow_examples.lite.examples.recommendation.ml_v2.configs.FeatureGroup)
  })
_sym_db.RegisterMessage(FeatureGroup)

InputConfig = _reflection.GeneratedProtocolMessageType('InputConfig', (_message.Message,), {
  'DESCRIPTOR' : _INPUTCONFIG,
  '__module__' : 'input_config_pb2'
  # @@protoc_insertion_point(class_scope:tensorflow_examples.lite.examples.recommendation.ml_v2.configs.InputConfig)
  })
_sym_db.RegisterMessage(InputConfig)


# @@protoc_insertion_point(module_scope)
# pyformat: enable



# Copyright 2020 Google Research. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Eval libraries."""
from absl import app
from absl import flags
from absl import logging
import tensorflow as tf

from tensorflow_examples.lite.model_maker.third_party.efficientdet import coco_metric
from tensorflow_examples.lite.model_maker.third_party.efficientdet import dataloader
from tensorflow_examples.lite.model_maker.third_party.efficientdet import hparams_config
from tensorflow_examples.lite.model_maker.third_party.efficientdet import utils
from tensorflow_examples.lite.model_maker.third_party.efficientdet.keras import anchors
from tensorflow_examples.lite.model_maker.third_party.efficientdet.keras import efficientdet_keras
from tensorflow_examples.lite.model_maker.third_party.efficientdet.keras import label_util
from tensorflow_examples.lite.model_maker.third_party.efficientdet.keras import postprocess
from tensorflow_examples.lite.model_maker.third_party.efficientdet.keras import util_keras

# Cloud TPU Cluster Resolvers
flags.DEFINE_string('tpu', None, 'The Cloud TPU name.')
flags.DEFINE_string('gcp_project', None, 'Project name.')
flags.DEFINE_string('tpu_zone', None, 'GCE zone name.')

flags.DEFINE_integer('eval_samples', None, 'Number of eval samples.')
flags.DEFINE_string('val_file_pattern', None,
                    'Glob for eval tfrecords, e.g. coco/val-*.tfrecord.')
flags.DEFINE_string('val_json_file', None,
                    'Groudtruth, e.g. annotations/instances_val2017.json.')
flags.DEFINE_string('model_name', 'efficientdet-d0', 'Model name to use.')
flags.DEFINE_string('model_dir', None, 'Location of the checkpoint to run.')
flags.DEFINE_integer('batch_size', 8, 'GLobal batch size.')
flags.DEFINE_string('hparams', '', 'Comma separated k=v pairs or a yaml file')
FLAGS = flags.FLAGS


def main(_):
  config = hparams_config.get_efficientdet_config(FLAGS.model_name)
  config.override(FLAGS.hparams)
  config.val_json_file = FLAGS.val_json_file
  config.nms_configs.max_nms_inputs = anchors.MAX_DETECTION_POINTS
  config.drop_remainder = False  # eval all examples w/o drop.
  config.image_size = utils.parse_image_size(config['image_size'])

  if config.strategy == 'tpu':
    tpu_cluster_resolver = tf.distribute.cluster_resolver.TPUClusterResolver(
        FLAGS.tpu, zone=FLAGS.tpu_zone, project=FLAGS.gcp_project)
    tf.config.experimental_connect_to_cluster(tpu_cluster_resolver)
    tf.tpu.experimental.initialize_tpu_system(tpu_cluster_resolver)
    ds_strategy = tf.distribute.TPUStrategy(tpu_cluster_resolver)
    logging.info('All devices: %s', tf.config.list_logical_devices('TPU'))
  elif config.strategy == 'gpus':
    ds_strategy = tf.distribute.MirroredStrategy()
    logging.info('All devices: %s', tf.config.list_physical_devices('GPU'))
  else:
    if tf.config.list_physical_devices('GPU'):
      ds_strategy = tf.distribute.OneDeviceStrategy('device:GPU:0')
    else:
      ds_strategy = tf.distribute.OneDeviceStrategy('device:CPU:0')

  with ds_strategy.scope():
    # Network
    model = efficientdet_keras.EfficientDetNet(config=config)
    model.build((None, *config.image_size, 3))
    util_keras.restore_ckpt(model,
                            tf.train.latest_checkpoint(FLAGS.model_dir),
                            config.moving_average_decay,
                            skip_mismatch=False)
    @tf.function
    def model_fn(images, labels):
      cls_outputs, box_outputs = model(images, training=False)
      detections = postprocess.generate_detections(config,
                                                   cls_outputs,
                                                   box_outputs,
                                                   labels['image_scales'],
                                                   labels['source_ids'])
      tf.numpy_function(evaluator.update_state,
                        [labels['groundtruth_data'],
                         postprocess.transform_detections(detections)], [])

    # Evaluator for AP calculation.
    label_map = label_util.get_label_map(config.label_map)
    evaluator = coco_metric.EvaluationMetric(
        filename=config.val_json_file, label_map=label_map)

    # dataset
    batch_size = FLAGS.batch_size   # global batch size.
    ds = dataloader.InputReader(
        FLAGS.val_file_pattern,
        is_training=False,
        max_instances_per_image=config.max_instances_per_image)(
            config, batch_size=batch_size)
    if FLAGS.eval_samples:
      ds = ds.take((FLAGS.eval_samples + batch_size - 1) // batch_size)
    ds = ds_strategy.experimental_distribute_dataset(ds)

    # evaluate all images.
    eval_samples = FLAGS.eval_samples or 5000
    pbar = tf.keras.utils.Progbar((eval_samples + batch_size - 1) // batch_size)
    for i, (images, labels) in enumerate(ds):
      ds_strategy.run(model_fn, (images, labels))
      pbar.update(i)

  # compute the final eval results.
  metrics = evaluator.result()
  metric_dict = {}
  for i, name in enumerate(evaluator.metric_names):
    metric_dict[name] = metrics[i]

  if label_map:
    for i, cid in enumerate(sorted(label_map.keys())):
      name = 'AP_/%s' % label_map[cid]
      metric_dict[name] = metrics[i + len(evaluator.metric_names)]
  print(FLAGS.model_name, metric_dict)


if __name__ == '__main__':
  flags.mark_flag_as_required('val_file_pattern')
  flags.mark_flag_as_required('model_dir')
  logging.set_verbosity(logging.ERROR)
  app.run(main)



# Copyright 2020 Google Research. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Model Builder for EfficientNet.

efficientnet-bx (x=0,1,2,3,4,5,6,7) checkpoints are located in:
  https://storage.googleapis.com/cloud-tpu-checkpoints/efficientnet/ckptsaug/efficientnet-bx.tar.gz
"""
import functools
import os
import re
from absl import logging
import numpy as np
import tensorflow as tf

from tensorflow_examples.lite.model_maker.third_party.efficientdet import utils
from tensorflow_examples.lite.model_maker.third_party.efficientdet.backbone import efficientnet_model


def efficientnet_params(model_name):
  """Get efficientnet params based on model name."""
  params_dict = {
      # (width_coefficient, depth_coefficient, resolution, dropout_rate)
      'efficientnet-b0': (1.0, 1.0, 224, 0.2),
      'efficientnet-b1': (1.0, 1.1, 240, 0.2),
      'efficientnet-b2': (1.1, 1.2, 260, 0.3),
      'efficientnet-b3': (1.2, 1.4, 300, 0.3),
      'efficientnet-b4': (1.4, 1.8, 380, 0.4),
      'efficientnet-b5': (1.6, 2.2, 456, 0.4),
      'efficientnet-b6': (1.8, 2.6, 528, 0.5),
      'efficientnet-b7': (2.0, 3.1, 600, 0.5),
      'efficientnet-b8': (2.2, 3.6, 672, 0.5),
      'efficientnet-l2': (4.3, 5.3, 800, 0.5),
  }
  return params_dict[model_name]


class BlockDecoder(object):
  """Block Decoder for readability."""

  def _decode_block_string(self, block_string):
    """Gets a block through a string notation of arguments."""
    assert isinstance(block_string, str)
    ops = block_string.split('_')
    options = {}
    for op in ops:
      splits = re.split(r'(\d.*)', op)
      if len(splits) >= 2:
        key, value = splits[:2]
        options[key] = value

    if 's' not in options or len(options['s']) != 2:
      raise ValueError('Strides options should be a pair of integers.')

    return efficientnet_model.BlockArgs(
        kernel_size=int(options['k']),
        num_repeat=int(options['r']),
        input_filters=int(options['i']),
        output_filters=int(options['o']),
        expand_ratio=int(options['e']),
        id_skip=('noskip' not in block_string),
        se_ratio=float(options['se']) if 'se' in options else None,
        strides=[int(options['s'][0]),
                 int(options['s'][1])],
        conv_type=int(options['c']) if 'c' in options else 0,
        fused_conv=int(options['f']) if 'f' in options else 0,
        super_pixel=int(options['p']) if 'p' in options else 0,
        condconv=('cc' in block_string))

  def _encode_block_string(self, block):
    """Encodes a block to a string."""
    args = [
        'r%d' % block.num_repeat,
        'k%d' % block.kernel_size,
        's%d%d' % (block.strides[0], block.strides[1]),
        'e%s' % block.expand_ratio,
        'i%d' % block.input_filters,
        'o%d' % block.output_filters,
        'c%d' % block.conv_type,
        'f%d' % block.fused_conv,
        'p%d' % block.super_pixel,
    ]
    if block.se_ratio > 0 and block.se_ratio <= 1:
      args.append('se%s' % block.se_ratio)
    if block.id_skip is False:  # pylint: disable=g-bool-id-comparison
      args.append('noskip')
    if block.condconv:
      args.append('cc')
    return '_'.join(args)

  def decode(self, string_list):
    """Decodes a list of string notations to specify blocks inside the network.

    Args:
      string_list: a list of strings, each string is a notation of block.

    Returns:
      A list of namedtuples to represent blocks arguments.
    """
    assert isinstance(string_list, list)
    blocks_args = []
    for block_string in string_list:
      blocks_args.append(self._decode_block_string(block_string))
    return blocks_args

  def encode(self, blocks_args):
    """Encodes a list of Blocks to a list of strings.

    Args:
      blocks_args: A list of namedtuples to represent blocks arguments.
    Returns:
      a list of strings, each string is a notation of block.
    """
    block_strings = []
    for block in blocks_args:
      block_strings.append(self._encode_block_string(block))
    return block_strings


def swish(features, use_native=True, use_hard=False):
  """Computes the Swish activation function.

  We provide three alternatives:
    - Native tf.nn.swish, use less memory during training than composable swish.
    - Quantization friendly hard swish.
    - A composable swish, equivalent to tf.nn.swish, but more general for
      finetuning and TF-Hub.

  Args:
    features: A `Tensor` representing preactivation values.
    use_native: Whether to use the native swish from tf.nn that uses a custom
      gradient to reduce memory usage, or to use customized swish that uses
      default TensorFlow gradient computation.
    use_hard: Whether to use quantization-friendly hard swish.

  Returns:
    The activation value.
  """
  if use_native and use_hard:
    raise ValueError('Cannot specify both use_native and use_hard.')

  if use_native:
    return tf.nn.swish(features)

  if use_hard:
    return features * tf.nn.relu6(features + np.float32(3)) * (1. / 6.)

  features = tf.convert_to_tensor(features, name='features')
  return features * tf.nn.sigmoid(features)


_DEFAULT_BLOCKS_ARGS = [
    'r1_k3_s11_e1_i32_o16_se0.25', 'r2_k3_s22_e6_i16_o24_se0.25',
    'r2_k5_s22_e6_i24_o40_se0.25', 'r3_k3_s22_e6_i40_o80_se0.25',
    'r3_k5_s11_e6_i80_o112_se0.25', 'r4_k5_s22_e6_i112_o192_se0.25',
    'r1_k3_s11_e6_i192_o320_se0.25',
]


def efficientnet(width_coefficient=None,
                 depth_coefficient=None,
                 dropout_rate=0.2,
                 survival_prob=0.8):
  """Creates a efficientnet model."""
  global_params = efficientnet_model.GlobalParams(
      blocks_args=_DEFAULT_BLOCKS_ARGS,
      batch_norm_momentum=0.99,
      batch_norm_epsilon=1e-3,
      dropout_rate=dropout_rate,
      survival_prob=survival_prob,
      data_format='channels_last',
      num_classes=1000,
      width_coefficient=width_coefficient,
      depth_coefficient=depth_coefficient,
      depth_divisor=8,
      min_depth=None,
      relu_fn=tf.nn.swish,
      # The default is TPU-specific batch norm.
      # The alternative is tf.layers.BatchNormalization.
      batch_norm=utils.TpuBatchNormalization,  # TPU-specific requirement.
      use_se=True,
      clip_projection_output=False)
  return global_params


def get_model_params(model_name, override_params):
  """Get the block args and global params for a given model."""
  if model_name.startswith('efficientnet'):
    width_coefficient, depth_coefficient, _, dropout_rate = (
        efficientnet_params(model_name))
    global_params = efficientnet(
        width_coefficient, depth_coefficient, dropout_rate)
  else:
    raise NotImplementedError('model name is not pre-defined: %s' % model_name)

  if override_params:
    # ValueError will be raised here if override_params has fields not included
    # in global_params.
    global_params = global_params._replace(**override_params)

  decoder = BlockDecoder()
  blocks_args = decoder.decode(global_params.blocks_args)

  logging.info('global_params= %s', global_params)
  return blocks_args, global_params


def build_model(images,
                model_name,
                training,
                override_params=None,
                model_dir=None,
                fine_tuning=False,
                features_only=False,
                pooled_features_only=False):
  """A helper function to create a model and return predicted logits.

  Args:
    images: input images tensor.
    model_name: string, the predefined model name.
    training: boolean, whether the model is constructed for training.
    override_params: A dictionary of params for overriding. Fields must exist in
      efficientnet_model.GlobalParams.
    model_dir: string, optional model dir for saving configs.
    fine_tuning: boolean, whether the model is used for finetuning.
    features_only: build the base feature network only (excluding final
      1x1 conv layer, global pooling, dropout and fc head).
    pooled_features_only: build the base network for features extraction (after
      1x1 conv layer and global pooling, but before dropout and fc head).

  Returns:
    logits: the logits tensor of classes.
    endpoints: the endpoints for each layer.

  Raises:
    When model_name specified an undefined model, raises NotImplementedError.
    When override_params has invalid fields, raises ValueError.
  """
  assert tf.is_tensor(images)
  assert not (features_only and pooled_features_only)

  # For backward compatibility.
  if override_params and override_params.get('drop_connect_rate', None):
    override_params['survival_prob'] = 1 - override_params['drop_connect_rate']

  if not training or fine_tuning:
    if not override_params:
      override_params = {}
    override_params['batch_norm'] = utils.BatchNormalization
    if fine_tuning:
      override_params['relu_fn'] = functools.partial(swish, use_native=False)
  blocks_args, global_params = get_model_params(model_name, override_params)

  if model_dir:
    param_file = os.path.join(model_dir, 'model_params.txt')
    if not tf.io.gfile.exists(param_file):
      if not tf.io.gfile.exists(model_dir):
        tf.io.gfile.makedirs(model_dir)
      with tf.io.gfile.GFile(param_file, 'w') as f:
        logging.info('writing to %s', param_file)
        f.write('model_name= %s\n\n' % model_name)
        f.write('global_params= %s\n\n' % str(global_params))
        f.write('blocks_args= %s\n\n' % str(blocks_args))

  model = efficientnet_model.Model(blocks_args, global_params, model_name)
  outputs = model(
      images,
      training=training,
      features_only=features_only,
      pooled_features_only=pooled_features_only)
  features, endpoints = outputs[0], outputs[1:]
  if features_only:
    features = tf.identity(features, 'features')
  elif pooled_features_only:
    features = tf.identity(features, 'pooled_features')
  else:
    features = tf.identity(features, 'logits')
  return features, endpoints


def build_model_base(images, model_name, training, override_params=None):
  """Create a base feature network and return the features before pooling.

  Args:
    images: input images tensor.
    model_name: string, the predefined model name.
    training: boolean, whether the model is constructed for training.
    override_params: A dictionary of params for overriding. Fields must exist in
      efficientnet_model.GlobalParams.

  Returns:
    features: base features before pooling.
    endpoints: the endpoints for each layer.

  Raises:
    When model_name specified an undefined model, raises NotImplementedError.
    When override_params has invalid fields, raises ValueError.
  """
  assert isinstance(images, tf.Tensor)
  # For backward compatibility.
  if override_params and override_params.get('drop_connect_rate', None):
    override_params['survival_prob'] = 1 - override_params['drop_connect_rate']

  blocks_args, global_params = get_model_params(model_name, override_params)

  model = efficientnet_model.Model(blocks_args, global_params, model_name)
  outputs = model(images, training=training, features_only=True)
  return outputs[0], outputs[1:]



# Copyright 2021 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Deprecated APIs."""

IMPORTS = {}

# pylint: disable=line-too-long
IMPORTS[''] = """
# Deprecated imports are kept for backward compatiblity and to be removed in
# future versions. Please refer to public APIs for replacement:
# https://www.tensorflow.org/lite/api_docs/python/tflite_model_maker
# pylint: disable=g-bad-import-order
from tensorflow_examples.lite.model_maker.core.data_util.image_dataloader import ImageClassifierDataLoader
from tensorflow_examples.lite.model_maker.core.export_format import ExportFormat
from tensorflow_examples.lite.model_maker.core.task import configs
# pylint: enable=g-bad-import-order
"""
# pylint: enable=line-too-long



# Copyright 2020 Google Research. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Common utils."""
import collections
import contextlib
import os
from typing import Text, Tuple, Union
from absl import logging
import gin
import numpy as np
import tensorflow.compat.v1 as tf
import tensorflow.compat.v2 as tf2
from tensorflow.python.eager import tape as tape_lib  # pylint:disable=g-direct-tensorflow-import
from tensorflow.python.tpu import tpu_function  # pylint:disable=g-direct-tensorflow-import
# pylint: disable=logging-format-interpolation


def srelu_fn(x):
  """Smooth relu: a smooth version of relu."""
  with tf.name_scope('srelu'):
    beta = tf.Variable(20.0, name='srelu_beta', dtype=tf.float32)**2
    beta = tf.cast(beta**2, x.dtype)
    safe_log = tf.math.log(tf.where(x > 0., beta * x + 1., tf.ones_like(x)))
    return tf.where((x > 0.), x - (1. / beta) * safe_log, tf.zeros_like(x))


def activation_fn(features: tf.Tensor, act_type: Text):
  """Customized non-linear activation type."""
  if act_type in ('silu', 'swish'):
    return tf.nn.swish(features)
  elif act_type == 'swish_native':
    return features * tf.sigmoid(features)
  elif act_type == 'hswish':
    return features * tf.nn.relu6(features + 3) / 6
  elif act_type == 'relu':
    return tf.nn.relu(features)
  elif act_type == 'relu6':
    return tf.nn.relu6(features)
  elif act_type == 'mish':
    return features * tf.math.tanh(tf.math.softplus(features))
  elif act_type == 'srelu':
    return srelu_fn(features)
  else:
    raise ValueError('Unsupported act_type {}'.format(act_type))


def cross_replica_mean(t, num_shards_per_group=None):
  """Calculates the average value of input tensor across TPU replicas."""
  num_shards = tpu_function.get_tpu_context().number_of_shards
  if not num_shards:
    return t

  if not num_shards_per_group:
    return tf.tpu.cross_replica_sum(t) / tf.cast(num_shards, t.dtype)

  group_assignment = None
  if num_shards_per_group > 1:
    if num_shards % num_shards_per_group != 0:
      raise ValueError('num_shards: %d mod shards_per_group: %d, should be 0' %
                       (num_shards, num_shards_per_group))
    num_groups = num_shards // num_shards_per_group
    group_assignment = [[
        x for x in range(num_shards) if x // num_shards_per_group == y
    ] for y in range(num_groups)]
  return tf.tpu.cross_replica_sum(t, group_assignment) / tf.cast(
      num_shards_per_group, t.dtype)


def get_ema_vars():
  """Get all exponential moving average (ema) variables."""
  ema_vars = (
      tf.trainable_variables() +
      tf.get_collection(tf.GraphKeys.MOVING_AVERAGE_VARIABLES))
  for v in tf.global_variables():
    # We maintain mva for batch norm moving mean and variance as well.
    if 'moving_mean' in v.name or 'moving_variance' in v.name:
      ema_vars.append(v)
  return list(set(ema_vars))


def get_ckpt_var_map(ckpt_path, ckpt_scope, var_scope, skip_mismatch=None):
  """Get a var map for restoring from pretrained checkpoints.

  Args:
    ckpt_path: string. A pretrained checkpoint path.
    ckpt_scope: string. Scope name for checkpoint variables.
    var_scope: string. Scope name for model variables.
    skip_mismatch: skip variables if shape mismatch.

  Returns:
    var_map: a dictionary from checkpoint name to model variables.
  """
  logging.info('Init model from checkpoint {}'.format(ckpt_path))
  if not ckpt_scope.endswith('/') or not var_scope.endswith('/'):
    raise ValueError('Please specific scope name ending with /')
  if ckpt_scope.startswith('/'):
    ckpt_scope = ckpt_scope[1:]
  if var_scope.startswith('/'):
    var_scope = var_scope[1:]

  var_map = {}
  # Get the list of vars to restore.
  model_vars = tf.get_collection(tf.GraphKeys.GLOBAL_VARIABLES, scope=var_scope)
  reader = tf.train.load_checkpoint(ckpt_path)
  ckpt_var_name_to_shape = reader.get_variable_to_shape_map()
  ckpt_var_names = set(reader.get_variable_to_shape_map().keys())

  if tf.distribute.get_replica_context():
    replica_id = tf.get_static_value(
        tf.distribute.get_replica_context().replica_id_in_sync_group)
  else:
    replica_id = 0

  for i, v in enumerate(model_vars):
    var_op_name = v.op.name

    if replica_id >= 1:
      var_op_name = ''.join(var_op_name.rsplit(f'/replica_{replica_id}', 1))

    if not var_op_name.startswith(var_scope):
      logging.info('skip {} -- does not match scope {}'.format(
          var_op_name, var_scope))
    ckpt_var = ckpt_scope + var_op_name[len(var_scope):]
    if 'global_step' in ckpt_var:
      continue

    if (ckpt_var not in ckpt_var_names and
        var_op_name.endswith('/ExponentialMovingAverage')):
      ckpt_var = ckpt_scope + var_op_name[:-len('/ExponentialMovingAverage')]

    if ckpt_var not in ckpt_var_names:
      if 'Momentum' in ckpt_var or 'RMSProp' in ckpt_var:
        # Skip optimizer variables.
        continue
      if skip_mismatch:
        logging.info('skip {} ({}) -- not in ckpt'.format(
            var_op_name, ckpt_var))
        continue
      raise ValueError('{} is not in ckpt {}'.format(v.op, ckpt_path))

    if v.shape != ckpt_var_name_to_shape[ckpt_var]:
      if skip_mismatch:
        logging.info('skip {} ({} vs {}) -- shape mismatch'.format(
            var_op_name, v.shape, ckpt_var_name_to_shape[ckpt_var]))
        continue
      raise ValueError('shape mismatch {} ({} vs {})'.format(
          var_op_name, v.shape, ckpt_var_name_to_shape[ckpt_var]))

    if i < 5:
      # Log the first few elements for sanity check.
      logging.info('Init {} from ckpt var {}'.format(var_op_name, ckpt_var))
    var_map[ckpt_var] = v

  return var_map


class TpuBatchNormalization(tf.keras.layers.BatchNormalization):
  """Cross replica batch normalization."""

  def __init__(self, fused=False, **kwargs):
    if not kwargs.get('name', None):
      kwargs['name'] = 'tpu_batch_normalization'
    if fused in (True, None):
      raise ValueError('TpuBatchNormalization does not support fused=True.')
    super().__init__(fused=fused, **kwargs)

  def _moments(self, inputs, reduction_axes, keep_dims, mask=None):
    """Compute the mean and variance: it overrides the original _moments."""
    shard_mean, shard_variance = super()._moments(
        inputs, reduction_axes, keep_dims=keep_dims, mask=mask)

    num_shards = tpu_function.get_tpu_context().number_of_shards or 1
    num_shards_per_group = min(32, num_shards)  # aggregate up to 32 cores.
    logging.info('TpuBatchNormalization with num_shards_per_group {}'.format(
        num_shards_per_group))
    if num_shards_per_group > 1:
      # Compute variance using: Var[X]= E[X^2] - E[X]^2.
      shard_square_of_mean = tf.math.square(shard_mean)
      shard_mean_of_square = shard_variance + shard_square_of_mean
      group_mean = cross_replica_mean(shard_mean, num_shards_per_group)
      group_mean_of_square = cross_replica_mean(shard_mean_of_square,
                                                num_shards_per_group)
      group_variance = group_mean_of_square - tf.math.square(group_mean)
      return (group_mean, group_variance)
    else:
      return (shard_mean, shard_variance)

  def call(self, inputs, mask=None, training=None):
    outputs = super().call(inputs, mask=mask, training=training)
    # A temporary hack for tf1 compatibility with keras batch norm.
    for u in self.updates:
      tf.add_to_collection(tf.GraphKeys.UPDATE_OPS, u)
    return outputs


class SyncBatchNormalization(tf.keras.layers.BatchNormalization):
  """Cross replica batch normalization."""

  def __init__(self, fused=False, **kwargs):
    if not kwargs.get('name', None):
      kwargs['name'] = 'tpu_batch_normalization'
    if fused in (True, None):
      raise ValueError('SyncBatchNormalization does not support fused=True.')
    super().__init__(fused=fused, **kwargs)

  def _moments(self, inputs, reduction_axes, keep_dims, mask=None):
    """Compute the mean and variance: it overrides the original _moments."""
    shard_mean, shard_variance = super()._moments(
        inputs, reduction_axes, keep_dims=keep_dims, mask=mask)

    replica_context = tf.distribute.get_replica_context()
    num_shards = replica_context.num_replicas_in_sync or 1

    if num_shards > 1:
      # Compute variance using: Var[X]= E[X^2] - E[X]^2.
      shard_square_of_mean = tf.math.square(shard_mean)
      shard_mean_of_square = shard_variance + shard_square_of_mean
      group_mean = replica_context.all_reduce(
          tf.distribute.ReduceOp.MEAN, shard_mean)
      group_mean_of_square = replica_context.all_reduce(
          tf.distribute.ReduceOp.MEAN, shard_mean_of_square)
      group_variance = group_mean_of_square - tf.math.square(group_mean)
      return (group_mean, group_variance)
    else:
      return (shard_mean, shard_variance)

  def call(self, inputs, mask=None, training=None):
    outputs = super().call(inputs, mask=mask, training=training)
    # A temporary hack for tf1 compatibility with keras batch norm.
    for u in self.updates:
      tf.add_to_collection(tf.GraphKeys.UPDATE_OPS, u)
    return outputs


class BatchNormalization(tf.keras.layers.BatchNormalization):
  """Fixed default name of BatchNormalization to match TpuBatchNormalization."""

  def __init__(self, **kwargs):
    if not kwargs.get('name', None):
      kwargs['name'] = 'tpu_batch_normalization'
    super().__init__(**kwargs)

  def call(self, inputs, mask=None, training=None):
    outputs = super().call(inputs, mask=mask, training=training)
    # A temporary hack for tf1 compatibility with keras batch norm.
    for u in self.updates:
      tf.add_to_collection(tf.GraphKeys.UPDATE_OPS, u)
    return outputs


def batch_norm_class(is_training, strategy=None):
  if is_training and strategy == 'tpu':
    return TpuBatchNormalization
  elif is_training and strategy == 'gpus':
    return SyncBatchNormalization
  else:
    return BatchNormalization

# A cache of variable scope to BatchNorm layer.
_BN_LAYER_CACHE = collections.defaultdict(dict)


@gin.configurable
def batch_normalization(inputs,
                        training=False,
                        strategy=None,
                        reuse_scope: bool = False,
                        **kwargs):
  """A wrapper for TpuBatchNormalization.

  Keras layers are incompatible with automatic tf scope reuse.

  Supports reuse of the exiting variable scope when a model is called multiple
  times. Otherwise, checkpoint weights would not be restored correctly.

  Args:
    inputs: Input to BatchNorm layer.
    training: Argument of Keras BatchNorm layer.
    strategy: Argument of Keras BatchNorm layer.
    reuse_scope: Whether to reuse existing layer in same scope.
    **kwargs: Arguments passed to Keras BatchNorm layer.

  Returns:
    Result of BatchNorm applied to inputs.
  """
  if reuse_scope:
    scope_name = tf.get_variable_scope().name
    bn_layer_cache = _BN_LAYER_CACHE[hash(inputs.graph)]
    if scope_name in bn_layer_cache:
      bn_layer = bn_layer_cache[scope_name]
      logging.info('Reusing variable scope %s', scope_name)
    else:
      bn_layer = batch_norm_class(training, strategy)(**kwargs)
      bn_layer_cache[scope_name] = bn_layer
  else:
    bn_layer = batch_norm_class(training, strategy)(**kwargs)
  return bn_layer(inputs, training=training)


def batch_norm_act(inputs,
                   is_training_bn: bool,
                   act_type: Union[Text, None],
                   init_zero: bool = False,
                   data_format: Text = 'channels_last',
                   momentum: float = 0.99,
                   epsilon: float = 1e-3,
                   strategy: Text = None,
                   name: Text = None,
                   batch_norm_trainable: bool = True):
  """Performs a batch normalization followed by a non-linear activation.

  Args:
    inputs: `Tensor` of shape `[batch, channels, ...]`.
    is_training_bn: `bool` for whether the model is training.
    act_type: non-linear relu function type. If None, omits the relu operation.
    init_zero: `bool` if True, initializes scale parameter of batch
      normalization with 0 instead of 1 (default).
    data_format: `str` either "channels_first" for `[batch, channels, height,
      width]` or "channels_last for `[batch, height, width, channels]`.
    momentum: `float`, momentume of batch norm.
    epsilon: `float`, small value for numerical stability.
    strategy: string to specify training strategy for TPU/GPU/CPU.
    name: the name of the batch normalization layer
    batch_norm_trainable: 'bool' if False, the batch statistics will not be
      updated.

  Returns:
    A normalized `Tensor` with the same `data_format`.
  """
  if init_zero:
    gamma_initializer = tf.zeros_initializer()
  else:
    gamma_initializer = tf.ones_initializer()

  if data_format == 'channels_first':
    axis = 1
  else:
    axis = 3

  inputs = batch_normalization(
      inputs=inputs,
      axis=axis,
      momentum=momentum,
      epsilon=epsilon,
      center=True,
      scale=True,
      trainable=batch_norm_trainable,
      training=is_training_bn,
      strategy=strategy,
      gamma_initializer=gamma_initializer,
      name=name)

  if act_type:
    inputs = activation_fn(inputs, act_type)
  return inputs


def drop_connect(inputs, is_training, survival_prob):
  """Drop the entire conv with given survival probability."""
  # "Deep Networks with Stochastic Depth", https://arxiv.org/pdf/1603.09382.pdf
  if not is_training:
    return inputs

  # Compute tensor.
  batch_size = tf.shape(inputs)[0]
  random_tensor = survival_prob
  random_tensor += tf.random.uniform([batch_size, 1, 1, 1], dtype=inputs.dtype)
  binary_tensor = tf.floor(random_tensor)
  # Unlike conventional way that multiply survival_prob at test time, here we
  # divide survival_prob at training time, such that no addition compute is
  # needed at test time.
  output = inputs / survival_prob * binary_tensor
  return output


def num_params_flops(readable_format=True):
  """Return number of parameters and flops."""
  nparams = np.sum(
      [np.prod(v.get_shape().as_list()) for v in tf.trainable_variables()])
  options = tf.profiler.ProfileOptionBuilder.float_operation()
  options['output'] = 'none'
  flops = tf.profiler.profile(
      tf.get_default_graph(), options=options).total_float_ops
  # We use flops to denote multiply-adds, which is counted as 2 ops in tfprof.
  flops = flops // 2
  if readable_format:
    nparams = float(nparams) * 1e-6
    flops = float(flops) * 1e-9
  return nparams, flops


conv_kernel_initializer = tf.initializers.variance_scaling()
dense_kernel_initializer = tf.initializers.variance_scaling()


class Pair(tuple):

  def __new__(cls, name, value):
    return super().__new__(cls, (name, value))

  def __init__(self, name, _):  # pylint: disable=super-init-not-called
    self.name = name


def scalar(name, tensor, is_tpu=True):
  """Stores a (name, Tensor) tuple in a custom collection."""
  logging.info('Adding scale summary {}'.format(Pair(name, tensor)))
  if is_tpu:
    tf.add_to_collection('scalar_summaries', Pair(name, tf.reduce_mean(tensor)))
  else:
    tf.summary.scalar(name, tf.reduce_mean(tensor))


def image(name, tensor, is_tpu=True):
  logging.info('Adding image summary {}'.format(Pair(name, tensor)))
  if is_tpu:
    tf.add_to_collection('image_summaries', Pair(name, tensor))
  else:
    tf.summary.image(name, tensor)


def get_tpu_host_call(global_step, params):
  """Get TPU host call for summaries."""
  scalar_summaries = tf.get_collection('scalar_summaries')
  if params['img_summary_steps']:
    image_summaries = tf.get_collection('image_summaries')
  else:
    image_summaries = []
  if not scalar_summaries and not image_summaries:
    return None  # No summaries to write.

  model_dir = params['model_dir']
  iterations_per_loop = params.get('iterations_per_loop', 100)
  img_steps = params['img_summary_steps']

  def host_call_fn(global_step, *args):
    """Training host call. Creates summaries for training metrics."""
    gs = global_step[0]
    with tf2.summary.create_file_writer(
        model_dir, max_queue=iterations_per_loop).as_default():
      with tf2.summary.record_if(True):
        for i, _ in enumerate(scalar_summaries):
          name = scalar_summaries[i][0]
          tensor = args[i][0]
          tf2.summary.scalar(name, tensor, step=gs)

      if img_steps:
        with tf2.summary.record_if(lambda: tf.math.equal(gs % img_steps, 0)):
          # Log images every 1k steps.
          for i, _ in enumerate(image_summaries):
            name = image_summaries[i][0]
            tensor = args[i + len(scalar_summaries)]
            tf2.summary.image(name, tensor, step=gs)

      return tf.summary.all_v2_summary_ops()

  reshaped_tensors = [tf.reshape(t, [1]) for _, t in scalar_summaries]
  reshaped_tensors += [t for _, t in image_summaries]
  global_step_t = tf.reshape(global_step, [1])
  return host_call_fn, [global_step_t] + reshaped_tensors


def archive_ckpt(ckpt_eval, ckpt_objective, ckpt_path):
  """Archive a checkpoint if the metric is better."""
  ckpt_dir, ckpt_name = os.path.split(ckpt_path)

  saved_objective_path = os.path.join(ckpt_dir, 'best_objective.txt')
  saved_objective = float('-inf')
  if tf.io.gfile.exists(saved_objective_path):
    with tf.io.gfile.GFile(saved_objective_path, 'r') as f:
      saved_objective = float(f.read())
  if saved_objective > ckpt_objective:
    logging.info('Ckpt {} is worse than {}'.format(ckpt_objective,
                                                   saved_objective))
    return False

  filenames = tf.io.gfile.glob(ckpt_path + '.*')
  if filenames is None:
    logging.info('No files to copy for checkpoint {}'.format(ckpt_path))
    return False

  # clear up the backup folder.
  backup_dir = os.path.join(ckpt_dir, 'backup')
  if tf.io.gfile.exists(backup_dir):
    tf.io.gfile.rmtree(backup_dir)

  # rename the old checkpoints to backup folder.
  dst_dir = os.path.join(ckpt_dir, 'archive')
  if tf.io.gfile.exists(dst_dir):
    logging.info('mv {} to {}'.format(dst_dir, backup_dir))
    tf.io.gfile.rename(dst_dir, backup_dir)

  # Write checkpoints.
  tf.io.gfile.makedirs(dst_dir)
  for f in filenames:
    dest = os.path.join(dst_dir, os.path.basename(f))
    tf.io.gfile.copy(f, dest, overwrite=True)
  ckpt_state = tf.train.generate_checkpoint_state_proto(
      dst_dir, model_checkpoint_path=os.path.join(dst_dir, ckpt_name))
  with tf.io.gfile.GFile(os.path.join(dst_dir, 'checkpoint'), 'w') as f:
    f.write(str(ckpt_state))
  with tf.io.gfile.GFile(os.path.join(dst_dir, 'best_eval.txt'), 'w') as f:
    f.write('%s' % ckpt_eval)

  # Update the best objective.
  with tf.io.gfile.GFile(saved_objective_path, 'w') as f:
    f.write('%f' % ckpt_objective)

  logging.info('Copying checkpoint {} to {}'.format(ckpt_path, dst_dir))
  return True


def parse_image_size(image_size: Union[Text, int, Tuple[int, int]]):
  """Parse the image size and return (height, width).

  Args:
    image_size: A integer, a tuple (H, W), or a string with HxW format.

  Returns:
    A tuple of integer (height, width).
  """
  if isinstance(image_size, int):
    # image_size is integer, with the same width and height.
    return (image_size, image_size)

  if isinstance(image_size, str):
    # image_size is a string with format WxH
    width, height = image_size.lower().split('x')
    return (int(height), int(width))

  if isinstance(image_size, tuple):
    return image_size

  raise ValueError('image_size must be an int, WxH string, or (height, width)'
                   'tuple. Was %r' % image_size)


def get_feat_sizes(image_size: Union[Text, int, Tuple[int, int]],
                   max_level: int):
  """Get feat widths and heights for all levels.

  Args:
    image_size: A integer, a tuple (H, W), or a string with HxW format.
    max_level: maximum feature level.

  Returns:
    feat_sizes: a list of tuples (height, width) for each level.
  """
  image_size = parse_image_size(image_size)
  feat_sizes = [{'height': image_size[0], 'width': image_size[1]}]
  feat_size = image_size
  for _ in range(1, max_level + 1):
    feat_size = ((feat_size[0] - 1) // 2 + 1, (feat_size[1] - 1) // 2 + 1)
    feat_sizes.append({'height': feat_size[0], 'width': feat_size[1]})
  return feat_sizes


def verify_feats_size(feats,
                      feat_sizes,
                      min_level,
                      max_level,
                      data_format='channels_last'):
  """Verify the feature map sizes."""
  expected_output_size = feat_sizes[min_level:max_level + 1]
  for cnt, size in enumerate(expected_output_size):
    h_id, w_id = (2, 3) if data_format == 'channels_first' else (1, 2)
    if feats[cnt].shape[h_id] != size['height']:
      raise ValueError(
          'feats[{}] has shape {} but its height should be {}.'
          '(input_height: {}, min_level: {}, max_level: {}.)'.format(
              cnt, feats[cnt].shape, size['height'], feat_sizes[0]['height'],
              min_level, max_level))
    if feats[cnt].shape[w_id] != size['width']:
      raise ValueError(
          'feats[{}] has shape {} but its width should be {}.'
          '(input_width: {}, min_level: {}, max_level: {}.)'.format(
              cnt, feats[cnt].shape, size['width'], feat_sizes[0]['width'],
              min_level, max_level))


def get_precision(strategy: str, mixed_precision: bool = False):
  """Get the precision policy for a given strategy."""
  if mixed_precision:
    if strategy == 'tpu':
      return 'mixed_bfloat16'

    if tf.config.list_physical_devices('GPU'):
      return 'mixed_float16'

    # TODO(fsx950223): Fix CPU float16 inference
    # https://github.com/google/automl/issues/504
    logging.warning('float16 is not supported for CPU, use float32 instead')
    return 'float32'

  return 'float32'


@contextlib.contextmanager
def float16_scope():
  """Scope class for float16."""

  def _custom_getter(getter, *args, **kwargs):
    """Returns a custom getter that methods must be called under."""
    cast_to_float16 = False
    requested_dtype = kwargs['dtype']
    if requested_dtype == tf.float16:
      kwargs['dtype'] = tf.float32
      cast_to_float16 = True
    var = getter(*args, **kwargs)
    if cast_to_float16:
      var = tf.cast(var, tf.float16)
    return var

  with tf.variable_scope('', custom_getter=_custom_getter) as varscope:
    yield varscope


def set_precision_policy(policy_name: Text = None):
  """Set precision policy according to the name.

  Args:
    policy_name: precision policy name, one of 'float32', 'mixed_float16',
      'mixed_bfloat16', or None.
  """
  if not policy_name:
    return

  assert policy_name in ('mixed_float16', 'mixed_bfloat16', 'float32')
  logging.info('use mixed precision policy name %s', policy_name)
  tf.compat.v1.keras.layers.enable_v2_dtype_behavior()
  # mixed_float16 training is not supported for now, so disable loss_scale.
  # float32 and mixed_bfloat16 do not need loss scale for training.
  policy = tf2.keras.mixed_precision.Policy(policy_name)
  tf2.keras.mixed_precision.set_global_policy(policy)


def build_model_with_precision(pp, mm, ii, *args, **kwargs):
  """Build model with its inputs/params for a specified precision context.

  This is highly specific to this codebase, and not intended to be general API.
  Advanced users only. DO NOT use it if you don't know what it does.
  NOTE: short argument names are intended to avoid conficts with kwargs.

  Args:
    pp: A string, precision policy name, such as "mixed_float16".
    mm: A function, for rmodel builder.
    ii: A tensor, for model inputs.
    *args: A list of model arguments.
    **kwargs: A dict, extra model parameters.

  Returns:
    the output of mm model.
  """
  if pp == 'mixed_bfloat16':
    set_precision_policy(pp)
    inputs = tf.cast(ii, tf.bfloat16)
    with tf.tpu.bfloat16_scope():
      outputs = mm(inputs, *args, **kwargs)
  elif pp == 'mixed_float16':
    set_precision_policy(pp)
    inputs = tf.cast(ii, tf.float16)
    with float16_scope():
      outputs = mm(inputs, *args, **kwargs)
  elif not pp or pp == 'float32':
    set_precision_policy(pp)
    outputs = mm(ii, *args, **kwargs)
  else:
    raise ValueError('Unknow precision name {}'.format(pp))

  # Users are responsible to convert the dtype of all outputs.
  return outputs


def _recompute_grad(f):
  """An eager-compatible version of recompute_grad.

  For f(*args, **kwargs), this supports gradients with respect to args or
  kwargs, but kwargs are currently only supported in eager-mode.
  Note that for keras layer and model objects, this is handled automatically.

  Warning: If `f` was originally a tf.keras Model or Layer object, `g` will not
  be able to access the member variables of that object, because `g` returns
  through the wrapper function `inner`.  When recomputing gradients through
  objects that inherit from keras, we suggest keeping a reference to the
  underlying object around for the purpose of accessing these variables.

  Args:
    f: function `f(*x)` that returns a `Tensor` or sequence of `Tensor` outputs.

  Returns:
   A function `g` that wraps `f`, but which recomputes `f` on the backwards
   pass of a gradient call.
  """

  @tf.custom_gradient
  def inner(*args, **kwargs):
    """Inner function closure for calculating gradients."""
    current_var_scope = tf.get_variable_scope()
    with tape_lib.stop_recording():
      result = f(*args, **kwargs)

    def grad_wrapper(*wrapper_args, **grad_kwargs):
      """Wrapper function to accomodate lack of kwargs in graph mode decorator."""

      @tf.custom_gradient
      def inner_recompute_grad(*dresult):
        """Nested custom gradient function for computing grads in reverse and forward mode autodiff."""
        # Gradient calculation for reverse mode autodiff.
        variables = grad_kwargs.get('variables')
        with tf.GradientTape() as t:
          id_args = tf.nest.map_structure(tf.identity, args)
          t.watch(id_args)
          if variables is not None:
            t.watch(variables)
          with tf.control_dependencies(dresult):
            with tf.variable_scope(current_var_scope):
              result = f(*id_args, **kwargs)
        kw_vars = []
        if variables is not None:
          kw_vars = list(variables)
        grads = t.gradient(
            result,
            list(id_args) + kw_vars,
            output_gradients=dresult,
            unconnected_gradients=tf.UnconnectedGradients.ZERO)

        def transpose(*t_args, **t_kwargs):
          """Gradient function calculation for forward mode autodiff."""
          # Just throw an error since gradients / activations are not stored on
          # tape for recompute.
          raise NotImplementedError(
              'recompute_grad tried to transpose grad of {}. '
              'Consider not using recompute_grad in forward mode'
              'autodiff'.format(f.__name__))

        return (grads[:len(id_args)], grads[len(id_args):]), transpose

      return inner_recompute_grad(*wrapper_args)

    return result, grad_wrapper

  return inner


def recompute_grad(recompute=False):
  """Decorator determine whether use gradient checkpoint."""

  def _wrapper(f):
    if recompute:
      return _recompute_grad(f)
    return f

  return _wrapper



# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Image classification demo code of Model Maker for TFLite."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os

from absl import app
from absl import flags
from absl import logging

import tensorflow as tf

from tflite_model_maker import image_classifier
from tflite_model_maker import model_spec
from tflite_model_maker.config import ExportFormat

FLAGS = flags.FLAGS


def define_flags():
  flags.DEFINE_string('export_dir', None,
                      'The directory to save exported files.')
  flags.DEFINE_string('spec', 'efficientnet_lite0',
                      'The image classifier to run.')
  flags.mark_flag_as_required('export_dir')


def download_demo_data(**kwargs):
  """Downloads demo data, and returns directory path."""
  data_dir = tf.keras.utils.get_file(
      fname='flower_photos.tgz',
      origin='https://storage.googleapis.com/download.tensorflow.org/example_images/flower_photos.tgz',
      extract=True,
      **kwargs)
  return os.path.join(os.path.dirname(data_dir), 'flower_photos')  # folder name


def run(data_dir, export_dir, spec='efficientnet_lite0', **kwargs):
  """Runs demo."""
  spec = model_spec.get(spec)
  data = image_classifier.DataLoader.from_folder(data_dir)
  train_data, rest_data = data.split(0.8)
  validation_data, test_data = rest_data.split(0.5)

  model = image_classifier.create(
      train_data, model_spec=spec, validation_data=validation_data, **kwargs)

  _, acc = model.evaluate(test_data)
  print('Test accuracy: %f' % acc)

  # Exports to TFLite and SavedModel, with label file.
  export_format = [
      ExportFormat.TFLITE,
      ExportFormat.SAVED_MODEL,
  ]
  model.export(export_dir, export_format=export_format)


def main(_):
  logging.set_verbosity(logging.INFO)
  data_dir = download_demo_data()
  export_dir = os.path.expanduser(FLAGS.export_dir)
  run(data_dir, export_dir, spec=FLAGS.spec)


if __name__ == '__main__':
  define_flags()
  app.run(main)



# Copyright 2022 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Util for converting ScaNN artifacts to on-device format."""
import os
from typing import NamedTuple, Optional, List, AnyStr

import numpy as np
import tensorflow as tf

from tensorflow_lite_support.scann_ondevice.cc.core import serialized_searcher_pb2
from scann.data_format import features_pb2
from scann.partitioning import partitioner_pb2
from scann.proto import centers_pb2
from scann.proto import hash_pb2
from scann.proto import scann_pb2
from tensorflow_lite_support.scann_ondevice.cc.python import index_builder

_DISTANCE_MAP = {
    'SquaredL2Distance': serialized_searcher_pb2.SQUARED_L2_DISTANCE,
    'DotProductDistance': serialized_searcher_pb2.DOT_PRODUCT,
}
_LOOKUP_TYPE_MAP = {
    hash_pb2.AsymmetricHasherConfig.FLOAT:
        serialized_searcher_pb2.AsymmetricHashingProto.FLOAT,
    hash_pb2.AsymmetricHasherConfig.INT8:
        serialized_searcher_pb2.AsymmetricHashingProto.INT8,
    hash_pb2.AsymmetricHasherConfig.INT16:
        serialized_searcher_pb2.AsymmetricHashingProto.INT16,
    hash_pb2.AsymmetricHasherConfig.INT8_LUT16:
        serialized_searcher_pb2.AsymmetricHashingProto.INT8_LUT16,
}


class OnDeviceArtifacts(NamedTuple):
  ondevice_config: serialized_searcher_pb2.ScannOnDeviceConfig
  hashed_dataset: Optional[np.ndarray]
  float_dataset: Optional[np.ndarray]
  partition_assignments: Optional[np.ndarray]


def get_distance_measure(
    distance_measure_str: str) -> serialized_searcher_pb2.DistanceMeasure:
  """Maps a distance measure string to a DistanceMeasure proto."""
  return _DISTANCE_MAP.get(distance_measure_str,
                           serialized_searcher_pb2.UNSPECIFIED)


def get_indexer(
    on_device_distance: serialized_searcher_pb2.DistanceMeasure,
    lookup_type: hash_pb2.AsymmetricHasherConfig.LookupType,
    ah_codebook: centers_pb2.CentersForAllSubspaces
) -> serialized_searcher_pb2.IndexerProto:
  """Helper util that builds an indexer for `convert_artifacts_to_leveldb`.

  Args:
    on_device_distance: DistanceMeasure used in NN search.
    lookup_type: Type of lookup table used for asymmetric distance NN search.
    ah_codebook: ScaNN AsymmetricHashing hasher.

  Returns:
    An IndexerProto.

  Raises:
    ValueError: If a subspace center in `ah_codebook` uses a feature type other
    than double or float.
  """
  indexer = serialized_searcher_pb2.IndexerProto()
  od_ah_codebook = indexer.asymmetric_hashing
  od_ah_codebook.query_distance = on_device_distance
  od_ah_codebook.lookup_type = _LOOKUP_TYPE_MAP[lookup_type]
  for subspace in ah_codebook.subspace_centers:
    od_subspace = od_ah_codebook.subspace.add()
    for center in subspace.center:
      entry = od_subspace.entry.add()
      feature_type = center.feature_type
      if feature_type == features_pb2.GenericFeatureVector.FeatureType.FLOAT:
        entry.dimension.extend(center.feature_value_float)
      elif feature_type == features_pb2.GenericFeatureVector.FeatureType.DOUBLE:
        entry.dimension.extend(center.feature_value_double)
      else:
        raise ValueError(
            f'ah_codebook has unsupported feature_type {feature_type}')
  return indexer


def get_partitioner(
    on_device_distance: serialized_searcher_pb2.DistanceMeasure,
    search_fraction: float,
    partition_centroids: partitioner_pb2.SerializedPartitioner
) -> serialized_searcher_pb2.PartitionerProto:
  """Helper util that builds a partitioner for `convert_artifacts_to_leveldb`.

  Args:
    on_device_distance: DistanceMeasure used in NN search.
    search_fraction: Fraction of partitions searched.
    partition_centroids: ScaNN partitioner.

  Returns:
    A PartitionerProto.
  """
  od_partitioner = serialized_searcher_pb2.PartitionerProto()
  od_partitioner.query_distance = on_device_distance
  od_partitioner.search_fraction = search_fraction
  for i, center in enumerate(
      partition_centroids.kmeans.kmeans_tree.root.centers):
    assert partition_centroids.kmeans.kmeans_tree.root.children[i].leaf_id == i
    leaf = od_partitioner.leaf.add()
    leaf.dimension.extend(center.dimension)
  return od_partitioner


def convert_serialized_to_on_device(serialized_path: str) -> OnDeviceArtifacts:
  """Converts ScaNN's serialized artifacts to on-device format.

  Args:
    serialized_path: Path to the dir that contains the ScaNN's artifacts.

  Returns:
    A named tuple containing the on-device ScannOnDeviceConfig as well as the
    database (maybe compressed) and partition assignment (if partitioning is
    enabled) in numpy format.
  """
  config_path = os.path.join(serialized_path, 'scann_config.pb')

  # Load ScannConfig
  with tf.io.gfile.GFile(config_path, 'rb') as config_pb:
    scann_config = scann_pb2.ScannConfig.FromString(config_pb.read())

  if scann_config.HasField('exact_reordering'):
    raise ValueError('exact_reordering is not supported on-device')

  on_device_config = serialized_searcher_pb2.ScannOnDeviceConfig()
  on_device_distance = get_distance_measure(
      scann_config.distance_measure.distance_measure)
  on_device_config.query_distance = on_device_distance

  # Indexed dataset parts
  ah_quantized_dataset = None
  float_dataset = None
  partition_assignments = None

  # Load AH centers
  if scann_config.HasField('hash'):
    with tf.io.gfile.GFile(
        os.path.join(serialized_path, 'ah_codebook.pb'), 'rb') as ah_pb:
      ah_codebook = centers_pb2.CentersForAllSubspaces.FromString(ah_pb.read())
      on_device_config.indexer.CopyFrom(
          get_indexer(on_device_distance,
                      scann_config.hash.asymmetric_hash.lookup_type,
                      ah_codebook))
    ah_quantized_dataset = np.load(
        os.path.join(serialized_path, 'hashed_dataset.npy'))

  # Load partition centroids
  if scann_config.HasField('partitioning'):
    with tf.io.gfile.GFile(
        os.path.join(serialized_path, 'serialized_partitioner.pb'),
        'rb') as partition_centroid_pb:
      partition_centroids = partitioner_pb2.SerializedPartitioner.FromString(
          partition_centroid_pb.read())
      on_device_config.partitioner.CopyFrom(
          get_partitioner(
              on_device_distance,
              (scann_config.partitioning.query_spilling.max_spill_centers /
               scann_config.partitioning.num_children), partition_centroids))
    partition_assignments = np.load(
        os.path.join(serialized_path, 'datapoint_to_token.npy'))

  # Load brute force datasets
  if scann_config.brute_force.fixed_point.enabled:
    raise NotImplementedError(
        'Fixed point int8 quantization is not yet supported on-device')
  elif scann_config.HasField('brute_force'):
    float_dataset = np.load(os.path.join(serialized_path, 'dataset.npy'))

  return OnDeviceArtifacts(on_device_config, ah_quantized_dataset,
                           float_dataset, partition_assignments)


def convert_artifacts_to_leveldb(output_file_path: str,
                                 metadata: List[AnyStr],
                                 userinfo: AnyStr,
                                 artifacts: OnDeviceArtifacts,
                                 compression: bool = True) -> None:
  """Converts artifacts to the index file.

  Raises exception if the input is invalid or failed to create the file.

  Args:
    output_file_path: Path to the levelDB index file.
    metadata: The metadata for each of the embeddings in the database. Passed in
      the same order as the embeddings in artifacts.
    userinfo: A special field in the index file that can be an arbitrary string
      supplied by the user.
    artifacts: Artifacts parsed by the convert_serialized_to_on_device
    compression: Whether to snappy compress the index file.
  """
  hashed_dataset = None
  float_dataset = None
  if artifacts.hashed_dataset is not None:
    embedding_dim = int(artifacts.hashed_dataset.shape[1])
    hashed_dataset = artifacts.hashed_dataset.reshape((-1,))
  if artifacts.float_dataset is not None:
    embedding_dim = int(artifacts.float_dataset.shape[1])
    float_dataset = artifacts.float_dataset.reshape((-1,))

  if len(artifacts.partition_assignments.shape) != 1:
    raise ValueError('Partition assignment array has to be 1D')

  # Raises exception if both hashed_dataset and float_dataset are not
  # None, or both are None
  serialized_index_file = index_builder.create_serialized_index_file(
      embedding_dim,
      artifacts.ondevice_config.SerializeToString(),
      userinfo,
      artifacts.partition_assignments,
      metadata,
      compression=compression,
      hashed_database=hashed_dataset,
      float_database=float_dataset)
  with tf.io.gfile.GFile(output_file_path, 'w') as f:
    f.write(serialized_index_file)



# Copyright 2021 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for Object Detection Dataloader."""

import abc
import hashlib
import io
import json
import os
import tempfile
from typing import Any, Collection, Dict, List, Sequence, Optional

import dataclasses
from lxml import etree
from PIL import JpegImagePlugin
import PIL.Image

import tensorflow as tf
import yaml

from tensorflow_examples.lite.model_maker.third_party.efficientdet.dataset import create_pascal_tfrecord
from tensorflow_examples.lite.model_maker.third_party.efficientdet.dataset import tfrecord_util

# A workaround to avoid a JPEG image being identified as MPO.
JpegImagePlugin._getmp = lambda: None  # pylint: disable=protected-access

# Suffix of the annotations json file name and the meta data file name.
ANN_JSON_FILE_SUFFIX = '_annotations.json'
META_DATA_FILE_SUFFIX = '_meta_data.yaml'


def _get_cache_dir_or_create(cache_dir: Optional[str]) -> str:
  """Gets the cache directory or creates it if not exists."""
  # TODO(b/183683348): Unifies with other tasks as well.
  # If `cache_dir` is None, a temporary folder will be created and will not be
  # removed automatically after training which makes it can be used later.
  if cache_dir is None:
    cache_dir = tempfile.mkdtemp()
  if not tf.io.gfile.exists(cache_dir):
    tf.io.gfile.makedirs(cache_dir)
  return cache_dir


def _get_dir_basename(dirname):
  """Gets the base name of the directory."""
  return os.path.basename(os.path.abspath(dirname))


def get_cache_prefix_filename_from_pascal(images_dir: str,
                                          annotations_dir: str,
                                          annotation_filenames: Optional[
                                              Collection[str]],
                                          num_shards: int = 10) -> str:
  """Gets the prefix of cached files from PASCAL VOC data.

  Args:
    images_dir: Path to directory that store raw images.
    annotations_dir: Path to the annotations directory.
    annotation_filenames: Collection of annotation filenames (strings) to be
      loaded. For instance, if there're 3 annotation files [0.xml, 1.xml, 2.xml]
      in `annotations_dir`, setting annotation_filenames=['0', '1'] makes this
      method only load [0.xml, 1.xml].
    num_shards: Number of shards for output file.

  Returns:
    The prefix of cached files.
  """
  hasher = hashlib.md5()
  hasher.update(_get_dir_basename(images_dir).encode('utf-8'))
  hasher.update(_get_dir_basename(annotations_dir).encode('utf-8'))
  if annotation_filenames:
    hasher.update(' '.join(sorted(annotation_filenames)).encode('utf-8'))
  hasher.update(str(num_shards).encode('utf-8'))
  return hasher.hexdigest()


def get_cache_prefix_filename_from_csv(csv_file: str, num_shards: int) -> str:
  """Gets the prefix of cached files from the csv file.

  Args:
    csv_file: Name of the csv file.
    num_shards: Number of shards for output file.

  Returns:
    The prefix of cached files.
  """
  hasher = hashlib.md5()
  hasher.update(os.path.basename(csv_file).encode('utf-8'))
  hasher.update(str(num_shards).encode('utf-8'))
  return hasher.hexdigest()


@dataclasses.dataclass(frozen=True)
class CacheFiles:
  """Cache files for object detection."""
  cache_prefix: str
  tfrecord_files: Sequence[str]
  meta_data_file: str
  annotations_json_file: Optional[str]


def get_cache_files(cache_dir: Optional[str],
                    cache_prefix_filename: str,
                    num_shards: int = 10) -> CacheFiles:
  """Creates an object of CacheFiles class.

  Args:
    cache_dir: The cache directory to save TFRecord, metadata and json file.
      When cache_dir is None, a temporary folder will be created and will not be
      removed automatically after training which makes it can be used later.
     cache_prefix_filename: The cache prefix filename.
     num_shards: Number of shards for output file.

  Returns:
    An object of CacheFiles class.
  """
  cache_dir = _get_cache_dir_or_create(cache_dir)
  # The cache prefix including the cache directory and the cache prefix
  # filename, e.g: '/tmp/cache/train'.
  cache_prefix = os.path.join(cache_dir, cache_prefix_filename)
  tf.compat.v1.logging.info(
      'Cache will be stored in %s with prefix filename %s. Cache_prefix is %s' %
      (cache_dir, cache_prefix_filename, cache_prefix))

  # Cached files including the TFRecord files, the annotations json file and
  # the meda data file.
  tfrecord_files = [
      cache_prefix + '-%05d-of-%05d.tfrecord' % (i, num_shards)
      for i in range(num_shards)
  ]
  annotations_json_file = cache_prefix + ANN_JSON_FILE_SUFFIX
  meta_data_file = cache_prefix + META_DATA_FILE_SUFFIX
  return CacheFiles(
      cache_prefix=cache_prefix,
      tfrecord_files=tuple(tfrecord_files),
      meta_data_file=meta_data_file,
      annotations_json_file=annotations_json_file)


def is_cached(cache_files: CacheFiles) -> bool:
  """Checks whether cache files are already cached."""
  # annotations_json_file is optional, thus we don't check whether it is cached.
  all_cached_files = list(
      cache_files.tfrecord_files) + [cache_files.meta_data_file]
  return all(tf.io.gfile.exists(path) for path in all_cached_files)


def is_all_cached(cache_files_collection: Collection[CacheFiles]) -> bool:
  """Checks whether a collection of cache files are all already cached."""
  return all(map(is_cached, cache_files_collection))


def get_cache_files_sequence(cache_dir: str, cache_prefix_filename: str,
                             set_prefixes: Collection[str],
                             num_shards: int) -> Sequence[CacheFiles]:
  """Gets a sequence of cache files.

  Args:
    cache_dir: The cache directory to save TFRecord, metadata and json file.
      When cache_dir is None, a temporary folder will be created and will not be
      removed automatically after training which makes it can be used later.
      cache_prefix_filename: The cache prefix filename.
    set_prefixes: Set prefix names for training, validation and test data. e.g.
      ['TRAIN', 'VAL', 'TEST'].
    num_shards: Number of shards for output file.

  Returns:
    A sequence of CachFiles objects mapping the set_prefixes.
  """
  cache_files_list = []
  for set_prefix in set_prefixes:
    prefix_filename = set_prefix.lower() + '_' + cache_prefix_filename
    cache_files = get_cache_files(cache_dir, prefix_filename, num_shards)
    cache_files_list.append(cache_files)
  return tuple(cache_files_list)


class CacheFilesWriter(abc.ABC):
  """CacheFilesWriter class to write the cached files."""

  def __init__(self,
               label_map: Dict[int, str],
               images_dir: Optional[str],
               num_shards: int = 10,
               max_num_images: Optional[int] = None,
               ignore_difficult_instances: bool = False) -> None:
    """Initializes CacheFilesWriter for object detector.

    Args:
      label_map: Dict, map label integer ids to string label names such as {1:
        'person', 2: 'notperson'}. 0 is the reserved key for `background` and
          doesn't need to be included in `label_map`. Label names can't be
          duplicated.
      images_dir: Path to directory that store raw images. If None, the image
        path is the path to Google Cloud Storage or the absolute path in the
        local machine.
      num_shards: Number of shards for the output file.
      max_num_images: Max number of images to process. If None, process all the
        images.
      ignore_difficult_instances: Whether to ignore difficult instances.
        `difficult` can be set inside `object` item in the annotation xml file.
    """
    self.label_map = label_map
    self.images_dir = images_dir if images_dir else ''
    self.num_shards = num_shards
    self.max_num_images = max_num_images
    self.ignore_difficult_instances = ignore_difficult_instances
    self.unique_id = create_pascal_tfrecord.UniqueId()

    self.label_name2id_dict = {'background': 0}
    for idx, name in self.label_map.items():
      self.label_name2id_dict[name] = idx

  def write_files(self, cache_files: CacheFiles, *args, **kwargs) -> None:
    """Writes TFRecord, meta_data and annotations json files.

    Args:
      cache_files: CacheFiles object including a list of TFRecord files, the
        annotations json file with COCO data format containing golden bounding
        boxes and the meda data yaml file to save the meta_data including data
        size and label_map.
      *args: Non-keyword of parameters used in the `_get_xml_dict` method.
      **kwargs: Keyword parameters used in the `_get_xml_dict` method.
    """
    writers = [
        tf.io.TFRecordWriter(path) for path in cache_files.tfrecord_files
    ]

    ann_json_dict = {'images': [], 'annotations': [], 'categories': []}
    for class_id, class_name in self.label_map.items():
      c = {'supercategory': 'none', 'id': class_id, 'name': class_name}
      ann_json_dict['categories'].append(c)

    # Writes tf.Example into TFRecord files.
    size = 0
    for idx, xml_dict in enumerate(self._get_xml_dict(*args, **kwargs)):
      if self.max_num_images and idx >= self.max_num_images:
        break
      if idx % 100 == 0:
        tf.compat.v1.logging.info('On image %d' % idx)
      tf_example = create_pascal_tfrecord.dict_to_tf_example(
          xml_dict,
          self.images_dir,
          self.label_name2id_dict,
          self.unique_id,
          ignore_difficult_instances=self.ignore_difficult_instances,
          ann_json_dict=ann_json_dict)
      writers[idx % self.num_shards].write(tf_example.SerializeToString())
      size = idx + 1

    for writer in writers:
      writer.close()

    # Writes meta_data into meta_data_file.
    meta_data = {'size': size, 'label_map': self.label_map}
    with tf.io.gfile.GFile(cache_files.meta_data_file, 'w') as f:
      yaml.dump(meta_data, f)

    # Writes ann_json_dict into annotations_json_file.
    with tf.io.gfile.GFile(cache_files.annotations_json_file, 'w') as f:
      json.dump(ann_json_dict, f, indent=2)

  @abc.abstractmethod
  def _get_xml_dict(self, *args, **kwargs) -> tf.train.Example:
    """Gets the dict holding PASCAL XML fields one by one."""
    raise NotImplementedError


class PascalVocCacheFilesWriter(CacheFilesWriter):
  """CacheFilesWriter class to write the cached files for Pascal Voc data."""

  def _get_xml_dict(
      self,
      annotations_dir: str,
      annotation_filenames: Optional[List[str]] = None) -> tf.train.Example:
    """Gets the tf example one by one from data with Pascal Voc format.

    Args:
      annotations_dir: Path to the annotations directory.
      annotation_filenames: Collection of annotation filenames (strings) to be
        loaded. For instance, if there're 3 annotation files [0.xml, 1.xml,
        2.xml] in `annotations_dir`, setting annotation_filenames=['0', '1']
        makes this method only load [0.xml, 1.xml].

    Yields:
      tf.train.Example
    """
    # Gets the paths to annotations.
    if annotation_filenames:
      ann_path_list = [
          os.path.join(annotations_dir, annotation + '.xml')
          for annotation in annotation_filenames
      ]
    else:
      ann_path_list = list(tf.io.gfile.glob(annotations_dir + r'/*.xml'))

    for ann_path in ann_path_list:
      with tf.io.gfile.GFile(ann_path, 'r') as fid:
        xml_str = fid.read()
      xml = etree.fromstring(xml_str)
      xml_dict = tfrecord_util.recursive_parse_xml_to_dict(xml)['annotation']
      yield xml_dict


def _get_xml_dict_from_csv_lines(images_dir: str, image_filename: str,
                                 lines: List[List[str]]) -> Dict[str, Any]:
  """Gets dict holding PASCAL VOC XML fields from the csv lines for an image."""
  image_path = os.path.join(images_dir, image_filename)
  with tf.io.gfile.GFile(image_path, 'rb') as fid:
    encoded_jpg = fid.read()
  encoded_jpg_io = io.BytesIO(encoded_jpg)
  image = PIL.Image.open(encoded_jpg_io)
  width, height = image.size

  xml_dict = {
      'size': {
          'width': width,
          'height': height
      },
      'filename': image_filename,
      'object': [],
  }

  for line in lines:
    label = line[2].strip()
    xmin, ymin = float(line[3]) * width, float(line[4]) * height
    xmax, ymax = float(line[7]) * width, float(line[8]) * height
    obj = {
        'name': label,
        'bndbox': {
            'xmin': xmin,
            'ymin': ymin,
            'xmax': xmax,
            'ymax': ymax,
            'name': label,
        },
        'difficult': 0,
        'truncated': 0,
        'pose': 'Unspecified',
    }
    xml_dict['object'].append(obj)

  return xml_dict


class CsvCacheFilesWriter(CacheFilesWriter):
  """DataLoader utilities to write the cached files for the csv file."""

  def _get_xml_dict(self, csv_lines: List[List[str]]) -> tf.train.Example:
    """Gets the tf example one by one from data with Pascal Voc format.

    Args:
      csv_lines: Lines in the csv files.

    Yields:
      tf.train.Example
    """
    image_dict = {}
    # Groups the line by image_path.
    for line in csv_lines:
      image_filename = line[1].strip()
      if image_filename not in image_dict:
        image_dict[image_filename] = []
      image_dict[image_filename].append(line)

    for image_filename, lines in image_dict.items():
      # Converts csv_lines for a certain image to dict holding PASCAL VOC XML
      # fields.
      xml_dict = _get_xml_dict_from_csv_lines(self.images_dir, image_filename,
                                              lines)
      yield xml_dict



# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""CLI tool for model maker."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import fire

from tensorflow_examples.lite.model_maker.core import compat
from tensorflow_examples.lite.model_maker.core.task import model_spec
from tensorflow_examples.lite.model_maker.demo import image_classification_demo
from tensorflow_examples.lite.model_maker.demo import question_answer_demo
from tensorflow_examples.lite.model_maker.demo import text_classification_demo

_IMAGE_MODELS = model_spec.IMAGE_CLASSIFICATION_MODELS
_TEXT_MODELS = model_spec.TEXT_CLASSIFICATION_MODELS
_QA_MODELS = model_spec.QUESTION_ANSWER_MODELS


class FormatDoc(object):
  """Decorator to format functionn doc with given parameters."""

  def __init__(self, *format_args, **format_kwargs):
    self.args = format_args
    self.kwargs = format_kwargs

  def __call__(self, fn):
    fn.__doc__ = fn.__doc__.format(*self.args, **self.kwargs)
    return fn


class ModelMakerCLI(object):
  """Model Maker Command Line Interface.

  Flags:
    --tf: int, version of TF behavior. Valid: [1, 2], default: 2.
  """

  def __init__(self, tf=2):
    compat.setup_tf_behavior(tf)

  @FormatDoc(MODELS=_IMAGE_MODELS)
  def image_classification(self,
                           data_dir,
                           export_dir,
                           spec='efficientnet_lite0',
                           **kwargs):
    """Run Image classification.

    Args:
      data_dir: str, input directory of training data. (required)
      export_dir: str, output directory to export files. (required)
      spec: str, model_name. Valid: {MODELS}, default: efficientnet_lite0.
      **kwargs: --epochs: int, epoch num to run. More: see `create` function.
    """
    # Convert types
    data_dir = str(data_dir)
    export_dir = str(export_dir)

    image_classification_demo.run(data_dir, export_dir, spec, **kwargs)

  @FormatDoc(MODELS=_TEXT_MODELS)
  def text_classification(self,
                          data_dir,
                          export_dir,
                          spec='mobilebert_classifier',
                          **kwargs):
    r"""Run text classification.

    Args:
      data_dir: str, input directory of training data. (required)
      export_dir: str, output directory to export files. (required)
      spec: str, model_name. Valid: {MODELS}, default: mobilebert_classifier.
      **kwargs: --epochs: int, epoch num to run. More: see `create` function.
    """
    # Convert types
    data_dir = str(data_dir)
    export_dir = str(export_dir)
    text_classification_demo.run(data_dir, export_dir, spec, **kwargs)

  @FormatDoc(MODELS=_QA_MODELS)
  def question_answer(self,
                      train_data_path,
                      validation_data_path,
                      export_dir,
                      spec='mobilebert_qa_squad',
                      **kwargs):
    r"""Run question answer.

    Args:
      train_data_path: str, input path of training data. (required)
      validation_data_path: str, input path of training data. (required)
      export_dir: str, output directory to export files. (required)
      spec: str, model_name. Valid: {MODELS}, default: mobilebert_qa.
      **kwargs: --epochs: int, epoch num to run. More: see `create` function.
    """
    # Convert types
    train_data_path = str(train_data_path)
    validation_data_path = str(validation_data_path)
    question_answer_demo.run(train_data_path, validation_data_path, export_dir,
                             spec, **kwargs)


def main():
  fire.Fire(ModelMakerCLI)


if __name__ == '__main__':
  main()



#   Copyright 2021 The TensorFlow Authors. All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Label encoder layer for on-device personalized recommendation model."""
import math
from typing import Dict

import tensorflow as tf

from tensorflow_examples.lite.model_maker.third_party.recommendation.ml.configs import input_config_pb2


class LabelEncoder(tf.keras.layers.Layer):
  """Layer to encode label feature.

  Currently only id-based label encoder is supported. With this encoder, label
  embedding layer will be created based on input config. And label embedding
  will be generated by feeding input label feature to label embedding layer.
  """

  def __init__(self,
               input_config: input_config_pb2.InputConfig):
    """Initialize LabelEncoder layer based on input config.

    Args:
      input_config: The input config containing label feature configuation,
        based on which label encoding could be generated.
    """
    super(LabelEncoder, self).__init__()
    self._input_config = input_config
    self._label_feature = self._input_config.label_feature
    if self._label_feature.feature_type != input_config_pb2.FeatureType.INT:
      raise ValueError('Currently only INT type is supported for label.')
    if not self.label_name:
      raise ValueError('label_feature name is expected.')
    self._label_embedding_layer = tf.keras.layers.Embedding(
        self._label_feature.vocab_size,
        self._label_feature.embedding_dim,
        embeddings_initializer=tf.keras.initializers.truncated_normal(
            mean=0.0,
            stddev=1.0 / math.sqrt(self._label_feature.embedding_dim)),
        mask_zero=True,
        name=self.label_name + 'embedding_layer')

  @property
  def label_name(self):
    return self._label_feature.feature_name

  def encode(self, input_label: tf.Tensor) -> tf.Tensor:
    """Generates label encoding with input label tensor.

    Args:
      input_label: Input label tensor, the shape of tensor should
        [num_label_items, 1].

    Returns:
      Label encoding.
    """
    if isinstance(input_label, tf.SparseTensor):
      input_label = tf.sparse.to_dense(input_label)
    label_embedding = self._label_embedding_layer(input_label)
    if tf.keras.backend.ndim(label_embedding) == 3:
      label_embedding = tf.squeeze(label_embedding)
    return label_embedding

  def call(self, inputs: Dict[str, tf.Tensor]):
    """Generate label encoding with input tentor dictionary.

    Args:
      inputs: Dictionary mapping feature name to input tensors.

    Returns:
      Label encoding.
    """
    input_label = inputs[self.label_name]
    return self.encode(input_label)
